@@ -1,25 +1,28 @@
-;;; erc-button.el --- A way of buttonizing certain things in ERC buffers
+;; erc-button.el --- A way of buttonizing certain things in ERC buffers
 
-;; Copyright (C) 2002  Mario Lang
+;; Copyright (C) 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004,
+;;   2006 Free Software Foundation, Inc.
 
 ;; Author: Mario Lang <mlang@delysid.org>
 ;; Keywords: irc, button, url, regexp
 ;; URL: http://www.emacswiki.org/cgi-bin/wiki.pl?ErcButton
 
-;; This file is free software; you can redistribute it and/or modify
+;; This file is part of GNU Emacs.
+
+;; GNU Emacs is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
 ;; the Free Software Foundation; either version 2, or (at your option)
 ;; any later version.
 
-;; This file is distributed in the hope that it will be useful,
+;; GNU Emacs is distributed in the hope that it will be useful,
 ;; but WITHOUT ANY WARRANTY; without even the implied warranty of
 ;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ;; GNU General Public License for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with GNU Emacs; see the file COPYING.  If not, write to
-;; the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-;; Boston, MA 02111-1307, USA.
+;; along with GNU Emacs; see the file COPYING.  If not, write to the
+;; Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
+;; Boston, MA 02110-1301, USA.
 
 ;;; Commentary:
 
@@ -40,6 +43,7 @@
 
 (require 'erc)
 (require 'wid-edit)
+(require 'erc-fill)
 
 ;;; Minor Mode
 
@@ -57,19 +61,33 @@
    (remove-hook 'erc-send-modify-hook 'erc-button-add-buttons)
    (remove-hook 'erc-complete-functions 'erc-button-next)))
 
+;; Make XEmacs use `erc-button-face'.
+(when (featurep 'xemacs)
+  (add-hook 'erc-mode-hook
+            (lambda () (set (make-local-variable 'widget-button-face) nil))))
+
 ;;; Variables
 
-(defconst erc-button-version "$Revision: 1.52 $"
+(defconst erc-button-version "$Revision: 1.71.2.2 $"
   "ERC button mode revision.")
 
-(defcustom erc-button-face 'bold
+(defface erc-button '((t (:bold t)))
+  "ERC button face."
+  :group 'erc-faces)
+
+(defcustom erc-button-face 'erc-button
   "Face used for highlighting buttons in ERC buffers.
 
-An button is a piece of text that you can activate by pressing
+A button is a piece of text that you can activate by pressing
 `RET' or `mouse-2' above it. See also `erc-button-keymap'."
   :type 'face
   :group 'erc-faces)
 
+(defcustom erc-button-nickname-face 'erc-nick-default-face
+  "Face used for ERC nickname buttons."
+  :type 'face
+  :group 'erc-faces)
+
 (defcustom erc-button-mouse-face 'highlight
   "Face used for mouse highlighting in ERC buffers.
 
@@ -78,7 +96,11 @@ above them."
   :type 'face
   :group 'erc-faces)
 
-(defcustom erc-button-url-regexp "\\(www\\.\\|\\(s?https?\\|ftp\\|file\\|gopher\\|news\\|telnet\\|wais\\|mailto\\):\\)\\(//[-a-zA-Z0-9_.]+:[0-9]*\\)?[-a-zA-Z0-9_=!?#$@~`%&*+\\/:;.,]+[-a-zA-Z0-9_=#$@~`%&*+\\/]"
+(defcustom erc-button-url-regexp
+  (concat "\\(www\\.\\|\\(s?https?\\|"
+          "ftp\\|file\\|gopher\\|news\\|telnet\\|wais\\|mailto\\):\\)"
+          "\\(//[-a-zA-Z0-9_.]+:[0-9]*\\)?"
+          "[-a-zA-Z0-9_=!?#$@~`%&*+\\/:;.,]+[-a-zA-Z0-9_=#$@~`%&*+\\/]")
   "Regular expression that matches URLs."
   :group 'erc-button
   :type 'regexp)
@@ -92,13 +114,10 @@ longer than `erc-fill-column'."
   :group 'erc-button
   :type '(choice integer boolean))
 
-(defcustom erc-button-buttonize-nicks 100
-  "Flag indicating whether nicks are buttonized or not.
-When the number of nicks on the channel is equal or smaller than the value
-of this variable, then nicks are buttonized.  When the number of nicks on
-the channel is larger, nicks are no longer buttonized."
+(defcustom erc-button-buttonize-nicks t
+  "Flag indicating whether nicks should be buttonized or not."
   :group 'erc-button
-  :type 'integer)
+  :type 'boolean)
 
 (defcustom erc-button-rfc-url "http://www.faqs.org/rfcs/rfc%s.html"
   "*URL used to browse rfc references.
@@ -117,17 +136,20 @@ the channel is larger, nicks are no longer buttonized."
   ;; a button, it makes no sense to optimize performance by
   ;; bytecompiling lambdas in this alist.  On the other hand, it makes
   ;; things hard to maintain.
-  '(("(\\(\\([^~\n \t@][^\n \t@]*\\)@\\([a-zA-Z0-9.:-]+\\)\\)" 1 t finger 2 3)
-    ("\\bInfo:[\"]\\([^\"]+\\)[\"]" 0 t Info-goto-node 1)
+  '(('nicknames 0 erc-button-buttonize-nicks erc-nick-popup 0)
+    (erc-button-url-regexp 0 t browse-url 0)
     ("<URL: *\\([^<> ]+\\) *>" 0 t browse-url 1)
+    ("(\\(\\([^~\n \t@][^\n \t@]*\\)@\\([a-zA-Z0-9.:-]+\\)\\)" 1 t finger 2 3)
+    ;; emacs internal
+    ("[`]\\([a-zA-Z][-a-zA-Z_0-9]+\\)[']" 1 t erc-button-describe-symbol 1)
+    ;; pseudo links
+    ("\\bInfo:[\"]\\([^\"]+\\)[\"]" 0 t Info-goto-node 1)
     ("\\b\\(Ward\\|Wiki\\|WardsWiki\\|TheWiki\\):\\([A-Z][a-z]+\\([A-Z][a-z]+\\)+\\)"
      0 t (lambda (page)
            (browse-url (concat "http://c2.com/cgi-bin/wiki?" page)))
      2)
     ("EmacsWiki:\\([A-Z][a-z]+\\([A-Z][a-z]+\\)+\\)" 0 t erc-browse-emacswiki 1)
     ("Lisp:\\([a-zA-Z.+-]+\\)" 0 t erc-browse-emacswiki-lisp 1)
-    (erc-button-url-regexp 0 t browse-url 0)
-    ('nicknames 0 erc-button-buttonize-nicks erc-nick-popup 0)
     ("\\bGoogle:\\([^ \t\n\r\f]+\\)"
      0 t (lambda (keywords)
            (browse-url (format erc-button-google-url keywords)))
@@ -136,25 +158,8 @@ the channel is larger, nicks are no longer buttonized."
      0 t (lambda (num)
            (browse-url (format erc-button-rfc-url num)))
      1)
-    ("\\s-\\(@\\([0-9][0-9][0-9]\\)\\)"
-     1 t (lambda (beats)
-           (let* ((seconds (- (* (string-to-number beats) 86.4)
-                              3600
-                              (- (car (current-time-zone)))))
-                  (hours (mod (floor seconds 3600) 24))
-                  (minutes (mod (round seconds 60) 60)))
-             (message (format "@%s is %d:%02d local time"
-                              beats hours minutes))))
-     2)
-    ("[`]\\([a-zA-Z][-a-zA-Z_0-9]+\\)[']"
-     1 t (lambda (symbol-name)
-           (let ((symbol (intern-soft symbol-name)))
-             (cond ((and symbol (fboundp symbol))
-                    (describe-function symbol))
-                   ((and symbol (boundp symbol))
-                    (describe-variable symbol))
-                   (t (apropos symbol-name)))))
-     1))
+    ;; other
+    ("\\s-\\(@\\([0-9][0-9][0-9]\\)\\)" 1 t erc-button-beats-to-time 2))
   "*Alist of regexps matching buttons in ERC buffers.
 Each entry has the form (REGEXP BUTTON FORM CALLBACK PAR...), where
 
@@ -168,7 +173,7 @@ REGEXP is the string matching text around the button or a symbol
   current server.
 
 BUTTON is the number of the regexp grouping actually matching the
-  button,  This is ignored if REGEXP is 'nickname.
+  button,  This is ignored if REGEXP is 'nicknames.
 
 FORM is a lisp expression which must eval to true for the button to
   be added,
@@ -179,16 +184,23 @@ CALLBACK is the function to call when the user push this button.
 
 PAR is a number of a regexp grouping whose text will be passed to
   CALLBACK.  There can be several PAR arguments.  If REGEXP is
-  'nickname, these are ignored, and CALLBACK will be called with
+  'nicknames, these are ignored, and CALLBACK will be called with
   the nickname matched as the argument."
   :group 'erc-button
-  :type '(repeat (list (choice regexp variable)
-                       (integer :tag "Button")
-                       (sexp :tag "Form")
-                       (function :tag "Callback")
-                       (repeat :tag "Par"
-                               :inline t
-                               (integer :tag "Regexp group")))))
+  :type '(repeat
+          (list :tag "Button"
+                (choice :tag "Matches"
+                        regexp
+                        (variable :tag "Variable containing regexp")
+                        (const :tag "Nicknames" 'nicknames))
+                (integer :tag "Number of the regexp section that matches")
+                (choice :tag "When to buttonize"
+                        (const :tag "Always" t)
+                        (sexp :tag "Only when this evaluates to non-nil"))
+                (function :tag "Function to call when button is pressed")
+                (repeat :tag "Sections of regexp to send to the function"
+                        :inline t
+                        (integer :tag "Regexp section number")))))
 
 (defcustom erc-emacswiki-url "http://www.emacswiki.org/cgi-bin/wiki.pl?"
   "*URL of the EmacsWiki Homepage."
@@ -207,6 +219,7 @@ PAR is a number of a regexp grouping whose text will be passed to
         (define-key map (kbd "<button2>") 'erc-button-click-button)
       (define-key map (kbd "<mouse-2>") 'erc-button-click-button))
     (define-key map (kbd "TAB") 'erc-button-next)
+    (define-key map (kbd "<backtab>") 'erc-button-previous)
     (set-keymap-parent map erc-mode-map)
     map)
   "Local keymap for ERC buttons.")
@@ -225,6 +238,7 @@ PAR is a number of a regexp grouping whose text will be passed to
     (modify-syntax-entry ?- "w" table)
     (modify-syntax-entry ?_ "w" table)
     (modify-syntax-entry ?| "w" table)
+    (modify-syntax-entry ?\\ "w" table)
     table)
   "Syntax table used when buttonizing messages.
 This syntax table should make all the legal nick characters word
@@ -278,7 +292,7 @@ specified by `erc-button-alist'."
                     (car bounds) (cdr bounds)))
         (if (erc-get-server-user word)
             (erc-button-add-button (car bounds) (cdr bounds)
-                                   fun (list word)))))))
+                                   fun t (list word)))))))
 
 (defun erc-button-add-buttons-1 (regexp entry)
   "Search through the buffer for matches to ENTRY and add buttons."
@@ -291,7 +305,7 @@ specified by `erc-button-alist'."
           (data (mapcar 'match-string (nthcdr 4 entry))))
       (when (or (eq t form)
                 (eval form))
-        (erc-button-add-button start end fun data regexp)))))
+        (erc-button-add-button start end fun nil data regexp)))))
 
 (defun erc-button-remove-old-buttons ()
   "Remove all existing buttons.
@@ -303,33 +317,37 @@ that `erc-button-add-button' adds, except for the face."
    '(erc-callback nil
                   erc-data nil
                   mouse-face nil
-                  local-map nil)))
+                  keymap nil)))
 
-(defun erc-button-add-button (from to fun &optional data regexp)
-  "Create a button between FROM and TO with callback FUN and data DATA."
+(defun erc-button-add-button (from to fun nick-p &optional data regexp)
+  "Create a button between FROM and TO with callback FUN and data DATA.
+NICK-P specifies if this is a nickname button.
+REGEXP is the regular expression which matched for this button."
   ;; Really nasty hack to <URL: > ise urls, and line-wrap them if
   ;; they're going to be wider than `erc-fill-column'.
   ;; This could be a lot cleaner, but it works for me -- lawrence.
-  (when (and erc-button-wrap-long-urls
-             (string= regexp erc-button-url-regexp)
-             (> (- to from) (if (numberp erc-button-wrap-long-urls)
-                                erc-button-wrap-long-urls
-                              erc-fill-column)))
-    (let ((erc-fill-column (if (numberp erc-button-wrap-long-urls)
-                               erc-button-wrap-long-urls
-                             erc-fill-column))
-          pos)
+  (let (fill-column)
+    (when (and erc-button-wrap-long-urls
+               (string= regexp erc-button-url-regexp)
+               (> (- to from)
+                  (setq fill-column (- (if (numberp erc-button-wrap-long-urls)
+                                           erc-button-wrap-long-urls
+                                         erc-fill-column)
+                                       (length erc-fill-prefix)))))
       (setq to (prog1 (point-marker) (insert ">"))
-            from   (prog2 (goto-char from) (point-marker) (insert "<URL: "))
-            pos (copy-marker from))
-      (while (> (- to pos) erc-fill-column)
-        (goto-char (+ pos erc-fill-column))
-        (insert "\n      ")             ; This ought to figure out
+            from (prog2 (goto-char from) (point-marker) (insert "<URL: ")))
+      (let ((pos (copy-marker from)))
+        (while (> (- to pos) fill-column)
+          (goto-char (+ pos fill-column))
+          (insert "\n" erc-fill-prefix) ; This ought to figure out
                                         ; what type of filling we're
                                         ; doing, and indent accordingly.
-        (move-marker pos (point)))))
-  (when erc-button-face
-    (erc-button-add-face from to erc-button-face))
+          (move-marker pos (point))))))
+  (if nick-p
+      (when erc-button-nickname-face
+        (erc-button-add-face from to erc-button-nickname-face))
+    (when erc-button-face
+      (erc-button-add-face from to erc-button-face)))
   (add-text-properties
    from to
    (nconc (and erc-button-mouse-face
@@ -339,6 +357,11 @@ that `erc-button-add-button' adds, except for the face."
           (list 'rear-nonsticky t)
           (and data (list 'erc-data data))))
   (widget-convert-button 'link from to :action 'erc-button-press-button
+                         :suppress-face t
+                         ;; Make XEmacs use our faces.
+                         :button-face (if nick-p
+                                          erc-button-nickname-face
+                                        erc-button-face)
                          ;; Make XEmacs behave with mouse-clicks, for
                          ;; some reason, widget stuff overrides the
                          ;; 'keymap text-property.
@@ -389,7 +412,7 @@ call it with the value of the `erc-data' text property."
     (unless fun
       (message "No button at point"))
     (when (and fun (symbolp fun) (not (fboundp fun)))
-      (error "Function %S is not bound"))
+      (error "Function %S is not bound" fun))
     (apply fun data)))
 
 (defun erc-button-next ()
@@ -408,6 +431,22 @@ call it with the value of the `erc-data' text property."
         (error "No next button"))
       t)))
 
+(defun erc-button-previous ()
+  "Go to the previous button in this buffer."
+  (interactive)
+  (let ((here (point)))
+    (when (< here (erc-beg-of-input-line))
+      (while (and (get-text-property here 'erc-callback)
+                  (not (= here (point-min))))
+        (setq here (1- here)))
+      (while (and (not (get-text-property here 'erc-callback))
+                  (not (= here (point-min))))
+        (setq here (1- here)))
+      (if (> here (point-min))
+          (goto-char here)
+        (error "No previous button"))
+      t)))
+
 (defun erc-browse-emacswiki (thing)
   "Browse to thing in the emacs-wiki."
   (browse-url (concat erc-emacswiki-url thing)))
@@ -450,9 +489,36 @@ Examples:
                                   erc-nick-popup-alist))
          (code (cdr (assoc action erc-nick-popup-alist))))
     (when code
-      (setq erc-active-buffer (current-buffer))
+      (erc-set-active-buffer (current-buffer))
       (eval code))))
 
+;;; Callback functions
+(defun erc-button-describe-symbol (symbol-name)
+  "Describe SYMBOL-NAME.
+Use `describe-function' for functions, `describe-variable' for variables,
+and `apropos' for other symbols."
+  (let ((symbol (intern-soft symbol-name)))
+    (cond ((and symbol (fboundp symbol))
+           (describe-function symbol))
+          ((and symbol (boundp symbol))
+           (describe-variable symbol))
+          (t (apropos symbol-name)))))
+
+(defun erc-button-beats-to-time (beats)
+  "Display BEATS in a readable time format."
+  (let* ((seconds (- (* (string-to-number beats) 86.4)
+                     3600
+                     (- (car (current-time-zone)))))
+         (hours (mod (floor seconds 3600) 24))
+         (minutes (mod (round seconds 60) 60)))
+    (message (format "@%s is %d:%02d local time"
+                     beats hours minutes))))
+
 (provide 'erc-button)
 
 ;;; erc-button.el ends here
+;; Local Variables:
+;; indent-tabs-mode: nil
+;; End:
+
+;; arch-tag: 7d23bed4-2f30-4273-a03f-d7a274c605c4
